from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _

User = get_user_model()


class CreatedUpdatedMixin(models.Model):
    class Meta:
        abstract = True

    created_at = models.DateTimeField(
        auto_now_add=True,
        help_text=_("The creation time, always in UTC timezone."),
    )
    updated_at = models.DateTimeField(
        auto_now=True,
        help_text=_("The last update time, always in UTC timezone."),
    )


class Category(CreatedUpdatedMixin):
    """Stores the categories of notifications"""

    title = models.CharField(max_length=30)
    icon = models.ImageField(upload_to="category/icons", null=True, blank=True)

    def __str__(self) -> str:
        return self.title

    class Meta:
        verbose_name = "Category"
        verbose_name_plural = "Categories"


class Notification(CreatedUpdatedMixin):
    """Stores the in app notifications of project"""

    class NotificationType(models.TextChoices):
        success = "success", _("Success")
        error = "error", _("Error")
        warning = "warning", _("Warning")
        info = "info", _("Information")

    class NotificationState(models.TextChoices):
        read = "read", _("Read")
        unread = "unread", _("Unread")
        deleted = "deleted", _("Deleted")

    title = models.CharField(max_length=265)
    message = models.TextField()
    category = models.ForeignKey(
        Category, on_delete=models.SET_NULL, null=True, blank=True
    )
    receivers = models.ManyToManyField(User, related_name="receivers_notifications")
    sender = models.ForeignKey(User, on_delete=models.CASCADE, null=True, blank=True)
    notification_type = models.CharField(
        choices=NotificationType.choices, max_length=20
    )
    state = models.CharField(
        choices=NotificationState.choices,
        max_length=20,
        default=NotificationState.unread,
    )

    def __str__(self) -> str:
        return self.title



import json

from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncJsonWebsocketConsumer

from notifications.models import Notification


class NotificationConsumer(AsyncJsonWebsocketConsumer):

    # Function to connect to the websocket
    async def connect(self):
        # Checking if the User is logged in
        self.user_id = self.scope["url_route"]["kwargs"]["user_id"]
        if not self.user_id:
            # Reject the connection
            await self.close()
        else:
            self.room_group_name = f"notifications_{self.user_id}"
            await self.channel_layer.group_add(self.room_group_name, self.channel_name)
            await self.accept()

    # Function to disconnet the Socket
    async def disconnect(self, close_code):
        await self.close()

    def get_async_notifications(self):
        self.notifications = list(
            Notification.objects.prefetch_related("receivers")
            .filter(
                receivers__id__in=[int(self.user_id)],
                state=Notification.NotificationState.unread,
            )
            .values("title", "message", "notification_type")
        )

    async def receive(self, text_data):
        text_data_json = json.loads(text_data)
        command = text_data_json["command"]
        # Send message to room group
        if command == "new_notification":
            content = text_data_json["content"]
            await self.channel_layer.group_send(
                self.room_group_name,
                {"type": "notify", "content": [content], "command": command},
            )
        elif command == "fetch_all_notifications":
            await database_sync_to_async(self.get_async_notifications)()
            await self.channel_layer.group_send(
                self.room_group_name,
                {"type": "notify", "content": self.notifications, "command": command},
            )

    # Custom Notify Function which can be called from Views or api to send message to the frontend
    async def notify(self, event):
        await self.send(
            text_data=json.dumps(
                {
                    "content": event["content"],
                }
            )
        )



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.contrib.auth import get_user_model
from django.db.models import F

from notifications.models import Category, Notification

User = get_user_model()


def send_notification(
    title: str,
    receivers: list[User],
    notification_type: str,
    sender: User = None,
    message: str = None,
    category: Category = None,
    real_time_notification: bool = True,
):
    """This function will send in app notification.

    Parameters:
    1. title : string
    2. receivers : List of User model instance
    3. sender : User model instance
    4. notification_type :
        a. success
        b. error
        c. warning
        d. info
    5. message : string
    6. category : Category Model instance
    7. notification_status :
        a. read
        b. unread
        c. deleted
    8. real_time_notification : bool (Set True if you want to implement Real Time Notifications)
    """
    if len(receivers) == 0:
        return "Please provide atleast one receiver."
    if notification_type not in list(
        dict(Notification.NotificationType.choices).keys()
    ):
        return f"Please provide valid notification type. It should be any of {list(dict(Notification.NotificationType.choices).keys())}."

    notification = Notification.objects.create(
        title=title,
        message=message,
        category=category,
        sender=sender,
        notification_type=notification_type,
        state=Notification.NotificationState.unread,
    )
    for receiver in receivers:
        notification.receivers.add(receiver)

    if real_time_notification:
        channel_layer = get_channel_layer()
        # Trigger message sent to group
        for receiver in receivers:
            async_to_sync(channel_layer.group_send)(
                f"notifications_{receiver.pk}",  # Group Name, Should always be string
                {
                    "type": "notify",  # Custom Function written in the consumers.py
                    "content": [
                        {
                            "title": title,
                            "message": message,
                            "notification_type": notification.get_notification_type_display(),
                            "category_type": category.title if category else "",
                            "sender_name": sender.username if sender else "",
                        }
                    ],
                    "command": "new_notification",
                },
            )
    return "Notification sent successfully."


def get_notifications(user, notification_type=None):
    if not notification_type:
        return (
            Notification.objects.prefetch_related("receivers")
            .filter(receivers=user)
            .annotate(
                sender_name=F("sender__first_name"), category_type=F("category__title")
            )
            .values(
                "title", "message", "notification_type", "sender_name", "category_type"
            )
        )

    if notification_type not in list(
        dict(Notification.NotificationType.choices).keys()
    ):
        return f"Please provide valid notification type. It should be any of {list(dict(Notification.NotificationType.choices).keys())}."
    else:
        return (
            Notification.objects.prefetch_related("receivers")
            .filter(receivers=user, notification_type=notification_type)
            .annotate(
                sender_name=F("sender__first_name"), category_type=F("category__title")
            )
            .values(
                "title", "message", "notification_type", "sender_name", "category_type"
            )
        )


def mark_as_read(user):
    Notification.objects.filter(
        receivers=user, state=Notification.NotificationState.unread
    ).update(state=Notification.NotificationState.read)

    return f"Notifications of user {user} marked as read."


def mark_as_unread(user):
    Notification.objects.filter(
        receivers=user, state=Notification.NotificationState.read
    ).update(state=Notification.NotificationState.unread)
    return f"Notifications of user {user} marked as unread."



import os

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from django.core.asgi import get_asgi_application

import notifications.routing

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "django_notifications.settings")

application = ProtocolTypeRouter(
    {
        "http": get_asgi_application(),
        "websocket": AuthMiddlewareStack(
            URLRouter(notifications.routing.websocket_urlpatterns)
        ),
    }
)



"""
To be used by pytests to override default database settings.
"""

from django_notifications.settings import *  # noqa

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",  # noqa
    }
}
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [("localhost", 6379)],
        },
    },
}
BROKER_BACKEND = "memory"



# Generated by Django 4.1.3 on 2023-07-03 10:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("notifications", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notification",
            name="state",
            field=models.CharField(
                choices=[
                    ("read", "Read"),
                    ("unread", "Unread"),
                    ("deleted", "Deleted"),
                ],
                default="unread",
                max_length=20,
            ),
        ),
    ]



# Generated by Django 4.1.3 on 2023-06-05 04:24

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The creation time, always in UTC timezone.",
                    ),
                ),
                (
                    "updated_at",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The last update time, always in UTC timezone.",
                    ),
                ),
                ("title", models.CharField(max_length=30)),
                (
                    "icon",
                    models.ImageField(
                        blank=True, null=True, upload_to="category/icons"
                    ),
                ),
            ],
            options={
                "verbose_name": "Category",
                "verbose_name_plural": "Categories",
            },
        ),
        migrations.CreateModel(
            name="Notification",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The creation time, always in UTC timezone.",
                    ),
                ),
                (
                    "updated_at",
                    models.DateTimeField(
                        auto_now=True,
                        help_text="The last update time, always in UTC timezone.",
                    ),
                ),
                ("title", models.CharField(max_length=265)),
                ("message", models.TextField()),
                (
                    "notification_type",
                    models.CharField(
                        choices=[
                            ("success", "Success"),
                            ("error", "Error"),
                            ("warning", "Warning"),
                            ("info", "Information"),
                        ],
                        max_length=20,
                    ),
                ),
                (
                    "state",
                    models.CharField(
                        choices=[
                            ("read", "Read"),
                            ("unread", "Unread"),
                            ("deleted", "Deleted"),
                        ],
                        max_length=20,
                    ),
                ),
                (
                    "category",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="notifications.category",
                    ),
                ),
                (
                    "receivers",
                    models.ManyToManyField(
                        related_name="receivers_notifications",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]
